/*
 *	Window functions. Slow slow slow slow mem hungry & slow...
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <string.h>

#include "hr_defines.h"
#include "hr_colors.h"
#include "hr_zones.h"
#include "hr_bobs.h"
#include "hr_mouse.h"
#include "hr_buttons.h"
#include "hr_general.h"
#include "hr_bitmaps.h"
#include "hr_input_handler.h"
#include "hr_fonts.h"
#include "hr_graphics.h"
#include "hr_screens.h"

struct hr_window_gadgets{
	int button_n;
	struct hr_button **button;
};


struct hr_window {
	int flags;
	struct hr_window *next;
	struct hr_window *prev;
	struct hr_color *colors;
	struct hr_bitmap *background;
	struct hr_bitmap *foreground;
	struct hr_zone *zone;
	struct hr_window_gadgets *gadgets;
	struct hr_menu *firstmenu;
};


/* Removes all windows from-first-to-wp from screen.
 * This must be called before moving,
 * resizing or in any way modifying windows. It is ok to modify contents
 * of frontmost window tho.*/
void hr_get_windows(struct hr_window *wp)
{
	struct hr_window *t;
	int width,height;
	int cx1,cy1,cx2,cy2,clip=1;
	if(hr.active_screen!=NULL){
		if(wp!=NULL){
			cx1=__clipx1;
			cy1=__clipy1;
			cx2=__clipx2;
			cy2=__clipy2;
			gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
			if(!__clip){
				clip=0;
				gl_enableclipping();
			}
			for(t=hr.active_screen->firstwindow;t!=(wp->next);t=t->next) {
				width=(t->zone->x2)-(t->zone->x1)+1;
				height=(t->zone->y2)-(t->zone->y1)+1;
				/* Save the window contents. */
				hr_delete_bitmap(t->foreground);
				t->foreground=hr_create_bitmap(malloc(width*height),
														 NULL,width,height,
														 0,GFX_COLOR);
				gl_getbox(t->zone->x1,t->zone->y1,
							 width,height,(t->foreground->data));
				/* Restore gfx under window, free mem.*/
				if(t->background!=NULL){
					gl_putbox(t->zone->x1,t->zone->y1,
								 width,height,
								 (t->background->data));
					hr_delete_bitmap(t->background);
					t->background=NULL;
				}
			}
			gl_setclippingwindow(cx1,cy1,cx2,cy2);
			if(!clip)
			  gl_disableclipping();
		}	
	}
}
	

/* Puts all windows back to their places, frees foregrounds. Call this
 * After moving, resizing or otherwise modifying any non-frontmost
 * window.*/
void hr_put_windows()
{
	struct hr_window *wp;
	int width,height;
	int cx1,cy1,cx2,cy2,clip=1;
	if(hr.active_screen!=NULL){
		if(hr.active_screen->firstwindow!=NULL){
			/**/
			cx1=__clipx1;
			cy1=__clipy1;
			cx2=__clipx2;
			cy2=__clipy2;
			gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
			if(!__clip){
				clip=0;
				gl_enableclipping();
			}
			/* Find last window */
			for(wp=hr.active_screen->firstwindow;wp->next!=NULL;wp=wp->next);
			/* In reverse order, "put" windows.*/
			for(;wp->prev!=NULL;wp=wp->prev){
				if(wp->foreground!=NULL){
					width=(wp->zone->x2)-(wp->zone->x1)+1;
					height=(wp->zone->y2)-(wp->zone->y1)+1;
					/* Save background */
					hr_delete_bitmap(wp->background);
					wp->background=hr_create_bitmap(malloc(width*height),
															  NULL,width,height,
															  0,GFX_COLOR);
					gl_getbox(wp->zone->x1,wp->zone->y1,
								 width,height,
								 wp->background->data);
					/* Restore foreground.*/
					if((width==(wp->foreground->width))&&
						(height==(wp->foreground->height)))
					  gl_putbox(wp->zone->x1,wp->zone->y1,
									width,height,wp->foreground->data);
					else
					  gl_putboxpart(wp->zone->x1,wp->zone->y1,
										 width,height,
										 (wp->foreground->width),
										 (wp->foreground->height),
										 (wp->foreground->data),0,0);
					/* Free foreground.*/
					hr_delete_bitmap(wp->foreground);
					wp->foreground=NULL;
				}
			}
			gl_setclippingwindow(cx1,cy1,cx2,cy2);
			if(!clip)
			  gl_disableclipping();
		}
	}
}

	
/* Redraws bevelbox, titlebac and stuff. Only works for frontmost window.*/
void hr_window_refresh(struct hr_window *wp)
{
	int bflags,l;
	int x1=0,x2=0,y1=0,y2=0,width=0,height=0,width2;
	struct hr_button *b;
	if(wp!=NULL){
		width=(wp->zone->x2)-(wp->zone->x1)+1;
		height=(wp->zone->y2)-(wp->zone->y1)+1;
		/* Remove windows on top of this one.*/
		hr_get_windows(wp->prev);
		/* Clear window area. */
		gl_fillbox((wp->zone->x1)+1,(wp->zone->y1)+1,
					  width-2,height-2,
					  hr_get_color_index(wp->colors,COLOR_BACKGROUND));
		/* Draw edge bevelbox.*/
		if(!(wp->flags & WINDOW_NOBORDER))
		  hr_bevelbox(wp->zone->x1,wp->zone->y1,
						  wp->zone->x2,wp->zone->y2,
						  wp->colors,NONE);
		/* Calc coordinates of innerbox.*/
		if(wp->flags & WINDOW_INNERBOX){
			x1=(wp->zone->x1)+2;
			x2=(wp->zone->x2)-2;
			y1=(wp->zone->y1)+2;
			y2=(wp->zone->y2)-2;
		}
		/* Set the window button positions and turn the buttons visible.*/
		if((wp->gadgets)!=NULL){
			if(wp->gadgets->button_n>0){
				width2=((width-3)/(wp->gadgets->button_n));
				/* Now put all the buttons in a row.*/
				for(l=0;l<(wp->gadgets->button_n);l++){
					b=(wp->gadgets->button)[l];
					if(b!=NULL) {
						b->zone->x1=(wp->zone->x1)+2+l*width2;
						b->zone->x2=(wp->zone->x1)+2+(l+1)*width2-2;
						b->zone->y1=(wp->zone->y1)+height-7-(b->font->height);
						if(wp->flags & WINDOW_INNERBOX){
							if((b->zone->y1)<(y2+2))
							  y2=(b->zone->y1)-2;
						}
						b->zone->y2=(wp->zone->y2)-2;
						/* Make the button visible */
						bflags=(b->flags &= ~(BUTTON_INVISIBLE));
						hr_update_button(b,UPDATE_FLAGS,&bflags);
					}
				}
			}
		}
		/* Draw inner bevelbox.*/
		if(wp->flags & WINDOW_INNERBOX)
		  hr_bevelbox(x1,y1,x2,y2,wp->colors,BUTTON_PRESSED);
		/* Restore windows on top of this one */
		hr_put_windows();
	}
}


/* Opens (creates) a window */
struct hr_window *hr_window_open(int x,int y,int w,int h,
											struct hr_color *colors,
											int flags)
{
	struct hr_window *wp=NULL;
	struct hr_button *b;
	int clip=1;
	if(hr.active_screen!=NULL){
		/* Redraw old default button without hilite-effect .*/
		if(hr.default_button!=NULL){
			b=hr.default_button;
			hr.default_button=NULL;
			hr_update_button(b,UPDATE_GFXONLY,NULL);
		}
		/* Create new window struct. It is put first (topmost) in the list.*/
		if((wp=malloc(sizeof(struct hr_window)))){
			wp->prev=NULL;
			wp->next=hr.active_screen->firstwindow;
			hr.active_screen->firstwindow=hr.active_screen->active_window=wp;
			if(wp->next!=NULL)
			  wp->next->prev=wp;
			/**/
			wp->colors=colors;
			wp->gadgets=NULL;
			wp->flags=flags;
			wp->firstmenu=NULL;
			wp->zone=hr_create_zone(x,y,x+w-1,y+h-1,ZONE_WINDOW);	
			/* Clipping window */
			if(flags & WINDOW_CLIPPING)
			  gl_setclippingwindow(wp->zone->x1,wp->zone->y1,
										  wp->zone->x2,wp->zone->y2);
			else
			  gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
			if(!__clip){
				gl_enableclipping();
				clip=0;
			}
			/* Store background */
			wp->foreground=NULL;
			wp->background=hr_create_bitmap(malloc(w*h),NULL,w,h,0,GFX_COLOR);
			gl_getbox(x,y,w,h,wp->background->data);			
			/* Draw it.*/
			hr_window_refresh(wp);
			if(!clip)
			  gl_disableclipping();
		}
	}
	return wp;
}


/* Closes a window. Since possible window buttons were (hopefully!)
 * initialized as INVISIBLE, the background is redrawn correctly.*/
void hr_window_close(struct hr_window *wp)
{
	struct hr_button *b;
	int l;
	int clip=1;
	if(hr.active_screen!=NULL){
		if(wp!=NULL) {
			gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
			if(!__clip){
				clip=0;
				gl_enableclipping();
			}
			/* Remove windows on top of wp.*/
			hr_get_windows(wp->prev);
			/* Restore gfx under window, free mem.*/
			if(wp->background!=NULL){
				gl_putbox(wp->zone->x1,wp->zone->y1,
							 (wp->zone->x2)-(wp->zone->x1)+1,
							 (wp->zone->y2)-(wp->zone->y1)+1,wp->background->data);
				hr_delete_bitmap(wp->background);
			}
			/* it is ok to free/delete NULLs...*/
			hr_delete_bitmap(wp->foreground);
			/* Delete window zone.*/
			hr_delete_zone(wp->zone);
			/* Delete the window buttons. It is ok to remove NULL buttons.*/
			if(wp->gadgets!=NULL){
				for(l=0;l<(wp->gadgets->button_n);l++)
				  hr_delete_button((wp->gadgets->button)[l]);
				free(wp->gadgets->button);
				free(wp->gadgets);
			}
			/* Set pointers.*/
			if(wp->next!=NULL)
			  wp->next->prev=wp->prev;
			if(wp->prev!=NULL)
			  wp->prev->next=wp->next;
			if(wp==hr.active_screen->firstwindow)
			  hr.active_screen->firstwindow=wp->next;
			if(wp==hr.active_screen->active_window)
			  hr.active_screen->active_window=wp->next;
			free(wp);
			/* Redraw windows on top of *wp if necessary */
			hr_put_windows();
			/* Clipping window.*/
			if(hr.active_screen->active_window!=NULL){
				if(hr.active_screen->active_window->flags & WINDOW_CLIPPING)
				  gl_setclippingwindow(hr.active_screen->active_window->zone->x1,
											  hr.active_screen->active_window->zone->y1,
											  hr.active_screen->active_window->zone->x2,
											  hr.active_screen->active_window->zone->y2);
			}
			if(!clip)
			  gl_disableclipping();
			/* Set default button - ie. found button that has BUTTON_DEFAULT_BUTTON flag
			 * set and that belongs to active_window.*/
			b=hr.firstbutton;
			while(b!=NULL){
				if((b->flags & BUTTON_DEFAULT_BUTTON) &&
					((b->parent_window)==hr.active_screen->active_window) &&
					(b->flags & BUTTON_ENABLED) &&
					!(b->flags & BUTTON_INVISIBLE) &&
					!(b->flags & BUTTON_NO_TAB_CYCLE))
				  {
					  hr.default_button=b;
					  hr_update_button(b,UPDATE_GFXONLY,NULL);
					  break;
				  }
				b=b->next;
			}
		}
	}
}


/* Moves a window */
void hr_window_move(struct hr_window *wp,int nx,int ny)
{
	if(wp!=NULL){
		hr_get_windows(wp);
		wp->zone->x1=nx;
		wp->zone->y1=ny;
		hr_put_windows();
	}
}


/* Resizes a window */
void hr_window_resize(struct hr_window *wp,int nwidth,int nheight)
{
	if(wp!=NULL){
		hr_get_windows(wp);
		wp->zone->x2=(wp->zone->x1)+nwidth-1;
		wp->zone->y2=(wp->zone->y1)+nheight-1;
		hr_put_windows();
	}
}


/* Moves a window to front.*/
void hr_window_to_front(struct hr_window *wp)
{
	struct hr_window *t;
	if(hr.active_screen!=NULL){
		if((wp!=NULL) && (wp!=hr.active_screen->firstwindow)){
			hr_get_windows(wp);
			for(t=hr.active_screen->firstwindow;t->next!=wp;t=t->next);
			t->next=t->next->next;
			t->next->prev=t->prev;
			wp->next=hr.active_screen->firstwindow;
			wp->prev=NULL;
			hr.active_screen->firstwindow=wp;
			hr_put_windows();
		}
	}
}
	
	
/* Moves a window to back */
void hr_window_to_back(struct hr_window *wp)
{
	struct hr_window *t;
	if(hr.active_screen!=NULL){
		if((wp!=NULL)&&(wp!=hr.active_screen->firstwindow)){
			hr_get_windows(wp);
			for(t=hr.active_screen->firstwindow;t->next!=wp;t=t->next);
			t->next=t->next->next;
			t->next->prev=t->prev;
			while(t->next!=NULL)
			  t=t->next;
			t->next=wp;
			wp->next=NULL;
			wp->prev=t;
			hr_put_windows();
		}
	}
}


/* Creates one window gadget struct and zeroes it.*/
struct hr_window_gadgets *hr_create_window_gadgets(int button_n)
{
	struct hr_window_gadgets *ggg=NULL;
	if((ggg=malloc(sizeof(struct hr_window_gadgets)))){
		ggg->button_n=button_n;
		if(button_n>0)
		  if(!(ggg->button=malloc(button_n*sizeof(struct hr_button *)))){
			  free(ggg);
			  return NULL;
		  }
	}
	return ggg;
}


/* Returns pointer to window under (x,y) */
struct hr_window *hr_get_window(int x,int y)
{
	struct hr_window *win;
	struct hr_zone *z;
	if(hr.active_screen!=NULL){
		if((z=hr_get_zone(hr.firstzone,x,y,ZONE_WINDOW))){	
			for(win=hr.active_screen->firstwindow;
				 win!=NULL;
				 win=win->next)
			  {
				  if(win->zone==z)
					 return win;
			  }
		}
	}
	return NULL;
}
